#define SLOW_MPI 0                               /* old version, obsolete */
#define MAX_out 1000        /* maximum lines for debugging/verbose output */
/* alpha dependency of terms (ToDo: and everything model specifique else) */ 
#ifndef USR_FILE   // may be defined outside
// * * * these files are used for models with external field * * *
//#define USR_FILE  "sample_bipartite_external_field.cpp"          
//#define USR_FILE  "sample_bipartite_external_field_j_j_dash.cpp" 
//#define USR_FILE  "sample_triangle_external_field_1.cpp"
//#define USR_FILE  "sample_triangle_external_field_2.cpp"
//#define USR_FILE  "sample_lc_j1j2j3.C"
//#define USR_FILE  "sample_lc_mf.C"
//#define USR_FILE  "sample_SrCuBO.C"
//#define USR_FILE "../documentation/stag_gen_ccm/sample_stag_spiral.C"
#endif
// Macros set outside:
//#define CFG_NoBra 1    // dont compute bra, ket + Eg only (RZinke, faster)
//#define TinyDelta 1e-5 // compute Delta-2*tinyDelta ... Delta+2*tinyDelta

/*

  /-------------------------------------------------\
  |                                                 |
  |                                                 |
  |     /=====\     /=====\     |\          /|      |
  |    /       \   /       \    | \        / |      |
  |    |           |            |  \      /  |      |
  |    |           |            |   \    /   |      |
  |    \       /   \       /    |    \  /    |      |
  |     \=====/     \=====/     |     \/     |      |
  |                                                 |
  |                                                 |
  \-------------------------------------------------/

  ===================================================

  Name:      CCM (solve.C)
  
  Written by: Damian J. J. Farnell

  Date:      Wed Mar 15 15:20:28 GMT 2000

  Updated:   Joerg Schulenburg 2004-2008

  ===================================================

  ---------------------------------------------------

  PROGRAM NAME: SOLVE.C 

  PURPOSE: Routine that solves saved CCM
           equations using newton-raphson
	   or direct iteration. Sublattice
           magnetisation determined and 
	   ground-state energy (+ derivatives).

  ---------------------------------------------------

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

// #include <iostream.h>
#include <stdlib.h>
#include <math.h>
#include <stdio.h>
#include <string.h>
#include <time.h>

#include "Constants.h"
#include "Externals.h"
#include "FunctionPrototypes.h"
#include "nr.h"
#include "nrutil.h"
#include "storage.h"
#include "f2c.h"
#include "eispackd.h"

#if (PARALLEL_FLAG>0)
#include <mpi.h>
#endif

// char gs_name[100]="";
// char input_name[200];
// char configurations[200];
int mapping[100]={0}; 
 /* The variable called "mapping" maps the 81 terms in the excited-state 
    equations to specific terms in the ground-state equations. 
    This is so that the "term" variable in the definition of the 
    delta[bonds][term] is correct. Again, seems to work quite well! */

/* Parameter array of size Params.bonds d1*H1+d2*H2+... */
// ToDo: JS091109:  fix exc-bug map81!=map62
// double (*delta)[62+1]=NULL;
// double (*dincr)[62+1]=NULL; // increment to next delta (derivation)
double delta[5][62+1]; // depends from usr_angle*usr_parameter=f(bond,term63)
double dincr[5][62+1]; // increment to next delta (derivation)
double delta81[5][81+1]; // fix exc-bug map81!=map62 

#ifdef USR_FILE // JS0812: was UseAlpha
// include user defined function usr_angle(bond,term13,*alpha,numalpha)
// Jun08 Joerg: includes USR_EXP_FILE in USR_FILE now (more simple)
#include USR_FILE  
#endif

// now macros which may be defined in USR_FILE (JS 2008/12)
/*  UseAlpha: maximum number parameter alpha used
 *        0 = no parameter alpha (default)
 *        1 = set parameter  alpha[0]    was alpha in old version
 *        2 = set parameters alpha[0..1] was alpha, beta in old version
 *        3 = set parameters alpha[0..2] was alpha,beta,gamma (old version)
 *      ToDo: better defined in USR_FILE, here only default
 */
#ifndef UseAlpha
#define UseAlpha      0
#endif
/* start angle: ToDo: we need <UseAlpha=1..3> angles here */
#ifndef StartAlpha
//#define StartAlpha    0.0
//#define StartAlpha    (M_PI/6.0)   /* triangle AF in external field 1 */
#define StartAlpha    (0.5*M_PI)     /* triangle AF in external field 2 */
#define StepAlpha     (0.001*M_PI)   /* initial step of alpha */
#endif
/*  SearchAlphaMin (meaning changed in v0904):
 *        0 = no search, alpha fixed (default)
 *        1 = raster 0..2Pi, 0..2Pi, 0..2Pi (for landscapes figures)
 *        2 = minimize energy by walking a 8 neighbour path
 *        3 = reserved for future
 */
#ifndef SearchAlphaMin     // do not minimize (default)
#define SearchAlphaMin 0
#endif
// set UseExpAlpha equal to 0 if you don't want to use user defined file
//        in find expectation values.
#ifndef UseExpAlpha
#define UseExpAlpha 0
#endif
// Jun08 Joerg: sample*expectation*.cpp added to sample*.cpp files
// Jun08 Joerg: alpha, beta, gamma replaced by alpha[0..2]
// ToDo: above declarations should be done in the USR_FILE (once!)

#define RandWalk 16 // 0 = no Random Walk Steps, 16 = max. 16 tests per step

/* -------------------------------------------------------------
 *  array of parameters depending on one variable parameter   
 *  sample: d1=cos(d), d2=sin(d) with H= d1*H1+d2*H2         
 */
static int PFLINE= __LINE__ + 1;   // this is only a localizator
// Joerg: depends from alpha only? can it depend from term (term13 or term63)?
// Joerg Jun08: remove term dependence
// double usr_parameter(int bonds, int term, double alpha)
#ifndef _USR_PARAMETER_   // may be defined in USR_FILE
double usr_parameter(int bonds, double alpha)
{
    /* add your model specifique parameter here! H=sum_i u(i,alpha)*H_i */
#if 1
#define _USR_PARAMETER_ "bond1=Delta"
  // square J1 J2 or xxzSquare
  //  tdelta * H1 + H2 + H3 + ...
  if (bonds==1) return alpha;
#elif 1
#define _USR_PARAMETER_ "bond1=bond2=Delta"
  // condition for triangular AF in external field: both model states
  if (bonds==1 || bonds==2) return alpha;
#else
#define _USR_PARAMETER_ "bond1=cos(Delta*Pi), bond2=sin(Delta*Pi), bond3=1"
  //  cos(alpha) * H1 + sin(alpha) * H2 + H3 + ...
  if (bonds==1) return cos(alpha*M_PI);
  if (bonds==2) return sin(alpha*M_PI); 
  if (bonds==3) return 1; 
#endif
  return 1.0;
}
#endif

double findEnergy(double *x);
void findExcitations(double *x, double *bra_x, double *alpha, double *ee);
void sublattice_term(int length, int (*configuration)[5], int *subTerm);
double findStagMag(double *x, double *bra_x, double *subLatticeMag);
void findBraState(double *x, double *bra_x);
void findBraStateRHS(double *rhs, double *x);
void findBraEquations(double **m, double *x);
int directIteration(int maximumNumIterations, int numConfigs, 
		    double tolerance, int *iterations, double *oldX);
int directIterationBraState(int maximumNumIterations, int numConfigs, 
			    double tolerance, int *iterations, 
			    double *oldX, double *oldBraX);
void subsetPrint(void); // define.C output subset rule
				
double *ex; // new code written 19 December 08 DJJF - direct iteration excited-state CCM
double *maxex; // new code written 19 December 08 DJJF - direct iteration excited-state CCM

void solve(void)
{
  int i, j, k;
  double *x, *bra_x; // ket and bra vector
  // index: 1..numConf, 0 used for code simplifications

  // char root_name[200], append[]=".dat";
  char output[200];
  //  char path[]="../Data/";
  //  char v_j2[]="_v_j2";
  //  int  **configuration, length;
  int iterations, iterations2, sum_it=0; /* ket, bra, allruns */
  double energy=0;
  //  double energy2;
  //  double specHeatA, specHeatC;
  int status=-1; // , stat;
  int sign = 1;

  double stagMag=0.0;
  /* we need it to find the inflection points of E(p) curves */
  double Energy1, Energy2;
  double dEnergy, ddEnergy;
  double stagMag1, stagMag2;
  double dStagMag, ddStagMag;
  double alpha[3]={0.0, 0.0, 0.0}; // alpha, beta, gamma
  double excitationEnergy=0.0;
  
  double start, stop, tdelta;
  double subLatticeMag[MAX_UNIT_CELL+1]={0.0, 0.0};
  double *g_exp; // extra space for GSE expectation

  time_t t0_solve, t0, t1;
  
#if UseAlpha
  int status_search=0;  // counting invalid states on searches
  double d_alpha[3],    // angle search accuracy or step
         oldE=1e9,      // last min found (last step to minimum)
         minE=1e9,      // min of neighbour step tests
         alpha_oldE[3]={0.0,0.0,0.0}, // angle for last minE found
         alpha_minE[3]={0.0,0.0,0.0}, // angle for minE of neighbour steps
         *x_minE=NULL, *x_oldE=NULL; // same for ket-vector

  // walking steps for min search
  // this walk is not random but tests vertices of a cube
  const int walkends[4]={ 1, 3, 9, 27}; // 3^UseAlpha - 1
  const int walkpath[27][3]={{ 0, 0, 0}, { 1, 0, 0}, {-1, 0, 0}, // 1d
                             { 1, 1, 0}, { 0, 1, 0}, {-1, 1, 0},
                             { 1,-1, 0}, { 0,-1, 0}, {-1,-1, 0}, // 2a
                             { 1, 1, 1}, { 0, 1, 1}, {-1, 1, 1},
                             { 1, 0, 1}, { 0, 0, 1}, {-1, 0, 1},
                             { 1,-1, 1}, { 0,-1, 1}, {-1,-1, 1},
                             { 1, 1,-1}, { 0, 1,-1}, {-1, 1,-1},
                             { 1, 0,-1}, { 0, 0,-1}, {-1, 0,-1},
                             { 1,-1,-1}, { 0,-1,-1}, {-1,-1,-1}}; // 3a
#if SearchAlphaMin
  x_minE= dvector(0,NumberConfigurations+1);
  x_oldE= dvector(0,NumberConfigurations+1);
#endif
#endif


  g_exp=      dvector(0,2*(Params.nExpect+1)); /* extra space for GSE */

  // configuration=imatrix(0,MAX_LSUBM+1,0,5);

  // now get the mappings right for the delta terms DJJF Jun08 Excitation
  //   map 1..81 to 1..62 [ to 1..13?]
  // see Externals.h map81[81+1] to 1..13 (remove mapping[81+1] to 1..62)
  
  for (i= 1;i<= 5;i++) mapping[i]= 2; // zz 1
  for (i= 6;i<= 6;i++) mapping[i]= 6; // z+ 2
  for (i= 7;i<=19;i++) mapping[i]= 8; // z- 3
  for (i=20;i<=20;i++) mapping[i]=14; // +z 4
  for (i=21;i<=33;i++) mapping[i]=16; // -z 5
  for (i=34;i<=36;i++) mapping[i]=22; // +- 6
  for (i=37;i<=39;i++) mapping[i]=24; // -+ 7 (no ++ 8)
  for (i=40;i<=73;i++) mapping[i]=27; // -- 9
  for (i=74;i<=74;i++) mapping[i]=54; // z  10
  for (i=75;i<=77;i++) mapping[i]=55; // -  11
  for (i=78;i<=81;i++) mapping[i]=58; // z^2 12 (no 13)
  for (i=1;i<=81;i++)
    if (map62[mapping[i]] != map81[i]) {
      printf("ERROR wrong mapping: map62[mapping[%d]]=%d != map81[%d]=%d\n",
              i, map62[mapping[i]], i, map81[i]);
    }
 
  //  solutionStrategy=1;
  //  momentum=0.7;         /* see Constants.h */

  if (!myid) subsetPrint(); // output subset rule (2nd time to be sure)
  time(&t0_solve);
  srand((int) t0_solve);

   // ToDo: remove _id000 (not needed if only one file)
// sprintf(input_name,    "%s.terms",  Control.name);
   sprintf(output,        "%s.output", Control.name);
// sprintf(configurations,"%s.configs",Control.name);
// sprintf(gs_name,       "%s_gs",     Control.name);

  if (myid==0)
    printf(" solve: myid=%3d output=%s\n", myid, output);
#ifdef _USR_PARAMETER_
  if (myid==0)
    printf("\n _USR_PARAMETER_(Delta): %s\n", _USR_PARAMETER_);
#endif

  // terms_fp=fopen(input_name,"r"); => opened in ccm.C now

  if (myid == 0) {
    fp2=fopen(output,"a");  
    if (!fp2) {
      printf("Cannot Open %s...\nExiting!!\n\n", output);
      ERR(1);
    }
  }

  // ground-state energy equation.

  // fp_gs=fopen(gs_name,"r");
  
  Energy2=1.0;    Energy1=2.0;   /* set initial values, no meaning */
  stagMag2=1.0;   stagMag1=2.0;

  x=      dvector(0,NumberConfigurations+1);
  bra_x=  dvector(0,NumberConfigurations+1);
  
  for(i=1;i<=NumberConfigurations;i++) { 
    x[i]=    0.0000001; // default=0.000001 ??? see xxzTriangle.ccm 
    bra_x[i]=0.000000001;
#if SearchAlphaMin
    x_minE[i]=x_oldE[i]=x[i];
#endif
  }
  // x[1]=0.1; // testwise, ToDo: failes, why?
  x[0]=    1.0; /* dummy index set to 1.0 for simplification, Joerg Sep05 */
  bra_x[0]=1.0; /* dummy index set to 1.0 for simplification, Joerg Sep05 */

//  if (solutionStrategy>1)
  {
    // new code written 19 December 08 DJJF - direct iteration excited-state CCM
    excitationEnergy=1.0;
    ex=      dvector(0,NumberExcitedConfigurations+5);
    maxex=   dvector(0,NumberExcitedConfigurations+5);
    for(i=1;i<=NumberExcitedConfigurations;i++)  {
       ex[i]=0.01; 
       maxex[i]=0.01;
    }
  }

  if (Control.output > -1 && myid == 0) {
    printf("\n ======== SOLVE CCM EQUATIONS: START ALGORITHM ========\n\n");
  }

  if (increment > 0.0) { start= minimum;  stop= maximum; }
    else               { start=-maximum;  stop=-minimum; }

   // Joerg Sep05: changing parameterfunction gives you more flexibility
  if (myid==0){ 
    printf("usr_parameter can be defined in "__FILE__" L%d\n",PFLINE);
    printf(" ... or define USR_FILE as include file in solve.C\n");
#ifdef USR_FILE
    printf(" ... here USR_FILE is defined as " USR_FILE "\n");
#endif
    printf("solutionStrategy= %d (%s)\n",
            solutionStrategy,
          ((solutionStrategy==1)?"newton-raphson":"direct iteration"));
    printf(" TOLERANCE= %.2e momentum= %.3f (see Constant.h)\n",
             TOLERANCE,      momentum);
  }
  /* array of parameters depending on one variable Parameter
   *  sample: d1=cos(d), d2=sin(d) with H= d1*H1+d2*H2
   */ 
  // delta = malloc((Params.bonds+1)*sizeof(double)*63);
  // dincr = malloc((Params.bonds+1)*sizeof(double)*63);
  // if (!delta || !dincr) {
  //   printf("input.C: delta = malloc() failed...\nExit!\n");
  //   exit(1);
  // }
  
  sign = ((increment<0.0) ? -1 : +1);

#if UseAlpha
    alpha_oldE[0] = alpha_minE[0] = alpha[0] =  StartAlpha;  // starting angle, 
    alpha_oldE[1] = alpha_minE[1] = alpha[1] =  0;
    alpha_oldE[2] = alpha_minE[2] = alpha[2] =  0;
#ifdef StartAlpha1
    alpha_oldE[1] = alpha_minE[1] = alpha[1] =  StartAlpha1;
#endif
#ifdef StartAlpha2
    alpha_oldE[2] = alpha_minE[2] = alpha[2] =  StartAlpha2;
#endif
    d_alpha[0] = StepAlpha;  // starting angle step
    d_alpha[1] = StepAlpha;  // starting angle step
    d_alpha[2] = StepAlpha;  // starting angle step
#endif

  // ToDo: load checkpoint if requested (tdelta + x + bra_x)
  if (!myid)
    printf(" ToDo: read possible checkpoint data on demand x + bra_x\n");

  // loop over different couplings J, go from known to unknown angle
  // also go around rounding errors for decimal dual conversion
  //   48bits deliver 48*l(2)/l(10) = 14.4 decimal digits
  //   we round up to 1e-12 for accumulated errors (typical loop 100)
  int iTinyD=0; // needed to trigger computation of dE and ddE
  for(tdelta=start; tdelta<=stop+1e-12;
                    tdelta+=fabs(increment), iTinyD++) {
#ifdef TinyDelta
   // used to calculate suszeptibility from derivations
   //   dE= (E(d+h)-E(d))/h
   //  ddE= (E(d+h)+E(d-h)-2E(d))/(2h)  lim(h=TinyDelta)=0
   // compute: d + {-2, -1, 0} * (h=TinyDelta), three steps for ddE
   double old_tdelta=tdelta;
   // Willert_Sep09: 3 steps instead of 5 (old), use iTinyD
   for (iTinyD=0; iTinyD<=2; iTinyD++) {
    tdelta = old_tdelta + (iTinyD - 2) * TinyDelta; // -2, -1, 0
#endif

    // D1(Delta)*H1 + D2(Delta)*H2 + D3(Delta)*H3 + ... , Joerg Oct05
    if (myid==0)
    printf("\nModel Parameter Delta = %8.5f\n", sign*tdelta);
// ToDo: replace by set_delta(in:tdelta,in:increment,out:delta,out:dincr)
{   int i;
    for (i=1; i<= Params.bonds; i++) {
      int ja=1; /* start range */ 
      for (j=1; j<=81; j++)
        delta81[i][j] = usr_parameter(i,sign* tdelta );
      for (j=1; j<=62; j++) {  // ToDo: use map62[1..62]
        delta[i][j] = usr_parameter(i,sign* tdelta );
        /* Joerg: dincr is only used for precondition (shift startvector) */
        dincr[i][j] = usr_parameter(i,sign*(tdelta+0.5*fabs(increment)))
                    - usr_parameter(i,sign*(tdelta-0.5*fabs(increment)));
        if (j>1 && delta[i][ja]!=delta[i][j] && myid==0 && Control.output>1) {
          printf(" Delta[bond=%d][term=%d...%d]= %6.4f\n",i,ja,j-1,delta[i][ja]);
          ja=j; /* next range */
        }
      } // j=1..62 (terms)
      if (myid==0 && Control.output>1)
        printf(" set Delta[bond=%d][term=%d...%d]= %6.4f\n",i,ja,j-1,delta[i][ja]);
    } // for i=1..Params.bounds
}
#if UseAlpha
    // find the the minimum E(angle), we assume that exact one minimum exist
    // increase accuracy of angle from 1e-1 to 1e-9 (interval)
    // ToDo: that seems not a good algorithm, use Newton method at least
    // go down with const steps, if it goes up, return with smaller steps
    // better algo, store last 5 lowest ergs(alpha[3]) or
    //    last steps(?) and MC in that region (random work?)
    // until 0902 min alpha[1] for fixed alpha[0], until min(alpha[0..1]) found
    //   rastering is slow
    // JS0902: random walk (alpha[0],alpha[1],alpha[2])
    //               range (d_alpha[0],d_alpha[1],d_alpha[2])*accuracyfactor
    // reduce d_alpha if no lower erg was found
    
    d_alpha[0] = fabs(StepAlpha); // start new minE(alpha) search, full range
    minE = oldE = 1.0e21; // biggest possible value
    for (int numAlphaSteps=0; // walkends[UseAlpha];
             fabs(d_alpha[0]) > 2e-9; numAlphaSteps++) {

#if   SearchAlphaMin == 1  // raster full alpha range
       // for rastering we need a strong defined startvector
       for(i=1;i<=NumberConfigurations;i++) x[i]=    0.000001; 
       for(i=1;i<=NumberConfigurations;i++) bra_x[i]=0.000000001; // ???
       // x[2]=x[1]=0.136; // triangle_field_2 better startvect? sub-4-4 
       // x[1]=0.1; // testwise, ToDo: failes, why?
       if (numAlphaSteps) alpha[0]+=d_alpha[0];
       if (alpha[0]>=alpha_oldE[0]+2*M_PI)  // overflow a[0]
         { alpha[0]= alpha_oldE[0]; alpha[1]+=d_alpha[1];
           if (!myid) printf("\n"); // for easier gnuplot.splot
           d_alpha[0]=1e-12; // alpha-loop end marker JS Oct09
         }
#if UseAlpha==2
       if (alpha[1]>=alpha_oldE[1]+2*M_PI) break;
#endif
       if (alpha[1]>=alpha_oldE[1]+2*M_PI)  // overflow a[1]
         { alpha[1]= alpha_oldE[1]; alpha[2]+=d_alpha[2]; } 
#if UseAlpha>=3
       if (alpha[2]>=alpha_oldE[2]+2*M_PI) break;
#endif
#elif SearchAlphaMin == 2  // walk along minimum path
       if (numAlphaSteps >= walkends[UseAlpha]) { // random walk
         if (!myid) {  // random() may daffer between tasks!
           alpha[0]=alpha_oldE[0]+d_alpha[0]*(random()*2.0/RAND_MAX-1.0);
           alpha[1]=alpha_oldE[1]+d_alpha[0]*(random()*2.0/RAND_MAX-1.0);
           alpha[2]=alpha_oldE[2]+d_alpha[0]*(random()*2.0/RAND_MAX-1.0);
         }
#if (PARALLEL_FLAG> 0)
         MPI_Bcast(alpha,3,MPI_DOUBLE,0,MPI_COMM_WORLD);
#endif  
       } else
       { int jj = numAlphaSteps % walkends[UseAlpha]; // deterministic walk
         alpha[0]=alpha_oldE[0]+d_alpha[0]*walkpath[jj][0];
         alpha[1]=alpha_oldE[1]+d_alpha[0]*walkpath[jj][1];
         alpha[2]=alpha_oldE[2]+d_alpha[0]*walkpath[jj][2];
         // for simplicity d_alpha is equal for all directions
       }
       // we always start with +(0,0,0), this is not optimal but easy
       for(i=1;i<=NumberConfigurations;i++) x[i]=x_oldE[i]; 
#endif

#if SearchAlphaMin
       if (Control.output>1) {
         printf( " set alpha                  a/Pi=");
         for (int j=0;j<UseAlpha;j++) printf(" % 10.8f", alpha[j]/M_PI);
         printf(" %5.0e aSteps= %4d\n", 
              fabs(d_alpha[0])/M_PI, numAlphaSteps );
       }
#endif

       for (i=1; i<= Params.bonds; i++) {
         for (j=1; j<=62; j++) {
           delta[i][j] = usr_angle(i,map62[j],alpha,UseAlpha )
                       * usr_parameter(i,sign* tdelta );
           dincr[i][j] = usr_parameter(i,sign*(tdelta+0.5*fabs(increment)))
                       - usr_parameter(i,sign*(tdelta-0.5*fabs(increment)));
           dincr[i][j]*=usr_angle(i,map62[j],alpha,UseAlpha );
           // dincr[i][j] = ??? 
         }
         for (j=1; j<=81; j++)
           delta81[i][j] = usr_angle(i,map81[j],alpha,UseAlpha )
                         * usr_parameter(i,sign* tdelta );
       }
#endif
    iterations=0;
    iterations2=0;
    time(&t0);      // start time
    
    // ToDo: set x to oldX (oldE) ???
    if (solutionStrategy==1) {
      
      // (a) newton-raphson ket state + LU decomp. bra state.
      // x(in,out) iterations(out), see num_methods.C
      status=mnewt((int) MAX_NUM_ITERATIONS, NumberConfigurations, 
		   (double) TOLERANCE, (double) TOLERANCE, &iterations, x);
      
    } else {
      
      // (b) direct iterations for both ket and bra states.

      status=directIteration((int) MAX_NUM_ITERATIONS,
             NumberConfigurations, (double) TOLERANCE, 
	     &iterations, x);
      // status=0 ok, status!=0 failure

    }
    time(&t1);  // status = 0 means OK, other values for errors

    sum_it += iterations; // sum for overall speed measuring

    if (iterations >= MAX_NUM_ITERATIONS || status) {
#if SearchAlphaMin>1
       status_search++; // count invalid search points
#endif
       for(i=1;i<=NumberConfigurations;i++) x[i]= 0.0;
       energy = 0.0;
       if (!status) status=-10;
    } else
      energy=findEnergy(x);

#if !UseAlpha
    if (Control.output>0 && myid==0) // print out later?
        printf(" %5d Eg= % 16.12f Eg.i100t[min]= % 6.2f\n",
           ((status)?-1:1)*iterations, energy, (t1-t0)/.6/iterations);
#else // UseAlpha
      // ToDo: reduce output for UsaAlpha<3
      if (Control.output>0 && myid==0) {
        int j;
        printf(" %5d Eg= % 16.12f", ((status)?-1:1)*iterations, energy);
        printf(" a/Pi=");
        for (j=0;j<UseAlpha;j++) printf(" % 10.8f", alpha[j]/M_PI);
#if SearchAlphaMin
        printf(" %5.0e aSteps= %4d", d_alpha[0]/M_PI, numAlphaSteps);
#endif
        printf("\n");
      }
#if SearchAlphaMin
      // minE is the minimum neighbour, oldE is the last position
      if (energy<minE && status==0 && iterations<MAX_NUM_ITERATIONS) { // JS0904
         minE = energy;      // minE of surrounding test steps
         alpha_minE[0] = alpha[0];
         alpha_minE[1] = alpha[1];
         alpha_minE[2] = alpha[2];
#if SearchAlphaMin == 1   // JS Oct09
         if (minE < oldE) oldE = minE;  // last taken step value
#endif
         for(i=1;i<=NumberConfigurations;i++) x_minE[i]=x[i]; 
         // skip rest of random walk tests if we have lower neighbour
#if SearchAlphaMin == 2
         if (numAlphaSteps+1 >= walkends[UseAlpha])
             numAlphaSteps += 10*RandWalk;
#endif
      } 
#if SearchAlphaMin == 2
      if (numAlphaSteps+1 >= walkends[UseAlpha]) { // ??? ToDo
        // stop if nothing was found
        if (minE < oldE) {
          if (fabs(minE - oldE) < TOLERANCE) break; // finish
          oldE = minE;       // last taken step value
          alpha_oldE[0] = alpha_minE[0];
          alpha_oldE[1] = alpha_minE[1];
          alpha_oldE[2] = alpha_minE[2];
          for(i=1;i<=NumberConfigurations;i++) x_oldE[i]=x_minE[i]; 
          numAlphaSteps = 0; // restart walk around cycle
          status_search=-9; // we are lucky enough
        } else {
          if (numAlphaSteps+1 >= walkends[UseAlpha] + RandWalk) {
            d_alpha[0] *= 0.5; // reduce steps
            numAlphaSteps = 0; // skip +(0,0,0)
          }
        }
      }
#endif
#else  // no SearchAlphaMin
      break; /* stop alpha search, d_alpha[0] = accuracy */
#endif

    } /* d_alpha[0], numAlphaSteps */
    
    // alpha=alpha_minE; energy=minE; // ???
#if SearchAlphaMin >= 1
    // set to min values for UseAlpha=0 outputs
    alpha[0] = alpha_minE[0] = alpha_oldE[0];
    alpha[1] = alpha_minE[1] = alpha_oldE[1];
    alpha[2] = alpha_minE[2] = alpha_oldE[2];
    energy = minE = oldE;
    for(i=1;i<=NumberConfigurations;i++) x[i]=x_minE[i]=x_oldE[i]; 
    if (energy>=1.0e21) {  // nothing found
      for(i=1;i<=NumberConfigurations;i++) { x[i]=0; x_oldE[i]=0.000001; }
      energy = minE = oldE = 0;
    }
    if (Control.output>-1 && myid==0) {
      printf(" Eg(alpha) %16.12f a/Pi=", minE);
      for (int j=0;j<UseAlpha;j++) printf(" % 10.8f", alpha[j]/M_PI);
      printf(" %5.0e\n", fabs(d_alpha[0])/M_PI); // accuracy Mar08, Feb09
    }
#endif  // SerchAlphaMin>=1
#endif  // UseAlpha

    // JS091021: add "|| status"
    if (iterations >= MAX_NUM_ITERATIONS || status) continue; // skip the rest
       // break;

    time(&t0);

#if !defined(CFG_NoBra)

    if (solutionStrategy==1) {
      
     // (a) LU decomp. bra state.

      // x(in), bra_x(out)
      findBraState(x, bra_x); // findBraStateRHS + findBraEq + LUdcmp
      iterations2=1; // ???

    } else {
      
      // (b) direct iterations for bra states.

      // ToDo: difference to initial state bra_x=0? JS09
      // x(in) bra_x(in(iterativ)?,out) iteratons2(out)
      status=directIterationBraState( (int)MAX_NUM_ITERATIONS, 
	     NumberConfigurations, (double)TOLERANCE, 
	     &iterations2, x, bra_x);
    }

    time(&t1);

    // (c) the (sub)lattice magnetisation.
      
    stagMag=findStagMag(x,bra_x,subLatticeMag);
	
    if (Control.output>0 && myid==0) // print out later?
        printf(" %5d M=  % 16.12f Mg.i100t[min]= % 6.2f\n",
           iterations2, stagMag, (t1-t0)/.6/iterations2);
    
	
    // now find the generalised expectations.
	
    if (Control.greatExpectations) findGenExp(x,bra_x,g_exp,alpha);

#else
    stagMag = 0.0;
#endif  // CFG_NoBra

    excitationEnergy=0.0;
    if (Control.excitations)
      findExcitations(x,bra_x,alpha,&excitationEnergy);

    // PRINT OUT INFORMATION.
    
    // we need at least two predecessing steps (iTinyD>=2)
    dEnergy  = ddEnergy  = 0;
    dStagMag = ddStagMag = 0;
    {
      double h = increment;
#ifdef TinyDelta
      h = TinyDelta;
#endif
      if (iTinyD>=1) dEnergy   = (energy -Energy1 )/(h);
      if (iTinyD>=1) dStagMag  = (stagMag-stagMag1)/(h);
      if (iTinyD>=2) ddEnergy  = (energy -2*Energy1 +Energy2 )/(h*h);
      if (iTinyD>=2) ddStagMag = (stagMag-2*stagMag1+stagMag2)/(h*h);
#ifdef TinyDelta
      if (iTinyD>1) // print only one of three steps
#endif
      if (Control.output > 0 && myid == 0) {
        if (iTinyD>=1) printf("Computational 1st Derivative Eg wrt delta = % f\n",dEnergy);
        if (iTinyD>=2) printf("Computational 2nd Derivative Eg wrt delta = % f\n",ddEnergy);
#if !defined(CFG_NoBra)
        if (iTinyD>=1) printf("Computational 1st Derivative  M wrt delta = % f\n",dStagMag);
        if (iTinyD>=2) printf("Computational 2nd Derivative  M wrt delta = % f\n",ddStagMag);
#endif
      }
    }    
    // = -d^2 E_g/ d Delta^2 via computational evaluation only.
    
    Energy2=Energy1;      Energy1=energy;   // save older values
    stagMag2=stagMag1;    stagMag1=stagMag;

    if (myid == 0) {
      printf("Delta = %6.4f : Eg = % 12.8f : M = % 12.8f : sum_it = %4d\n",
	     sign*tdelta, energy, stagMag, sum_it /* iterations */);
      if (Control.output>0)
      printf("               -dEg = % 12.8f : Heat = % 12.8f\n"
             "ket[1]= % 12.9f [2]= % 12.9f\n"
             "bra[1]= % 12.9f [2]= % 12.9f\n",
	      -dEnergy, ddEnergy,
                 x[1],     x[(1%NumberConfigurations)+1],
             bra_x[1], bra_x[(1%NumberConfigurations)+1]);
      if (Control.output>0 && UnitCell.number>1) {
	for(k=1;k<=UnitCell.number;k++)
	  printf("Mag[%d]= % 12.9f\n", k, subLatticeMag[k]);
      }
#if UseAlpha
      printf("Alpha= % 13.9f %5.0e\n",alpha_minE[0]/M_PI,
                                    fabs(d_alpha[0])/M_PI);
#endif
#if UseAlpha >= 2
      printf(" Beta= % 13.9f\n",alpha_minE[1]/M_PI);
#endif
#if UseAlpha >= 3
      printf("Gamma= % 13.9f\n",alpha_minE[2]/M_PI);
#endif
      
      if (Control.excitations && myid==0)  // DJJF Jun08
         printf("Excited-State Energy =%16.10f\n", excitationEnergy);

      if (Control.greatExpectations>0) {
        double sum_g_exp=0;
        for (k=1;k<=Params.nExpect;k++) {
          printf("Generalised Expectation Value [%d] = % 12.8f\n", k, g_exp[k]);
          sum_g_exp += g_exp[k];
        }
        // Jun08 Joerg: old was g_exp[1]+...+g_exp[3] which can be uninitialized
        printf("Lattice Magnetisation Triangular AF=%12.8f\n",sum_g_exp);
      }
      if (Control.output > 0) {
        int i, j0, j1, j2; double minX=9999, maxX=-1;
        for(j0=j1=j2=0,i=1;i<=NumberConfigurations;i++) {
          if (fabs(x[i])< minX) minX=fabs(x[i]);
          if (fabs(x[i])> maxX) maxX=fabs(x[i]);
          if (     x[i] <-1e-9) j0++;
          if (fabs(x[i])< 1e-9) j2++; 
          if ( bra_x[i] <-1e-9) j1++;
        }
        printf("\nmaxX = max(|x[i]|) = % 8.2e\n", maxX);
        printf(  "minX = min(|x[i]|) = % 8.2e\n", minX);
        if (minX/maxX<1e-12) { // set a more practical zero border
          minX=1e-12;
          printf("minX/maxX<1e-12, set new minX=1e-12\n");
        }
        printf("number coefficients (x<-t,|x|<t,bra_x<-t,t=1e-9)   "
               " = %d %d %d  NC= %d\n", j0, j2, j1, NumberConfigurations);
        for(j0=j1=j2=0,i=1;i<=NumberConfigurations;i++) {
          if (     x[i] <-10*minX) j0++;
          if (fabs(x[i])< 20*minX) j2++; 
          if ( bra_x[i] <-10*minX) j1++;
        }
        printf("number coefficients (x<-t,|x|<t,bra_x<-t,t=10*minX)"
               " = %d %d %d  NC= %d\n", j0, j2, j1, NumberConfigurations);

        if (j0<100 && Control.output>1) { // lot of output
            for(j0=0,i=1;i<=NumberConfigurations;i++) {
              if (x[i]<0) {
                if(j0< 1000) printf(" %d",i);
                if(j0==1000) printf("\n");
                j0++;
              }
            }
            printf("\n");
            for(j0=0,i=1;i<=NumberConfigurations;i++) {
              if (bra_x[i]<0) {
                if(j0< 1000) printf(" %d",i);
                if(j0==1000) printf("\n");
                j0++;
              }
            }
            printf("\n");
          }
        }
        if (Control.output > 1) {
          int i; double sum=0,sumb=0,prod=0;
          for(i=1;i<=NumberConfigurations;i++) {
            if (i<Control.output*50)
              printf(" i x[i] bra[i] %4d % .12f % .12f\n",
                       i,x[i],bra_x[i]);
            if (i==Control.output*20) printf("...\n");
            sum+=x[i]*x[i]; sumb+=bra_x[i]*bra_x[i]; prod+=x[i]*bra_x[i];
          }
          printf(" x.x bra.bra x.bra  % .9f % .9f % .9f\n",sum,sumb,prod);
        }
      }

#ifdef TinyDelta
      if (iTinyD>1) // print only one of three steps
#endif
      if (myid == 0) {
        static int headline=0;
        if (!headline) {
          fprintf(fp2,"# delta           Eg             stagMag           "
                   "-dE              ddE         x[1]            it it_bra");
#if UseAlpha
          for (int i=0; i<UseAlpha; i++) fprintf(fp2," alpha[%d]/Pi", i);
          fprintf(fp2," precision");
#endif
          fprintf(fp2,"\n");
          headline=1; // print once only! JS0902
        }
        fprintf(fp2,"%16.12f %16.12f %16.12f %16.12f %16.12f %16.12f %5d %5d",
  	  sign*tdelta, energy, stagMag, -dEnergy, ddEnergy, x[1], 
          iterations, iterations2);
#if UseAlpha
        for (int i=0; i<UseAlpha; i++) fprintf(fp2," %16.12f",alpha[i]/M_PI);
        fprintf(fp2," %5.0e",fabs(d_alpha[0])/M_PI);
#endif
      if(UnitCell.number>1) {
	for(k=1;k<=UnitCell.number;k++)
	  fprintf(fp2," %16.12f",subLatticeMag[k]);
      }
      double sum1=0.0;
      if (Control.greatExpectations) {
        for (k=1;k<=Params.nExpect;k++) sum1+=g_exp[k];
        for (k=1;k<=Params.nExpect;k++) fprintf(fp2," %16.12f", g_exp[k]);
      }
      fprintf(fp2," %16.12f",sum1); // for triangular AD lattice mag only
      if (Control.excitations) { 
        fprintf(fp2," excErg= % 16.12f", excitationEnergy);
      }
      fprintf(fp2,"\n");
      fflush(fp2);
    } // myid == 0

#ifdef TinyDelta
   } // tdelta +/- TinyDelta
   tdelta=old_tdelta;
#endif

   if (!myid) {
     time(&t1);
     if (t1-t0_solve>NumberConfigurations/1000000*60+60) { // allow writing 8MB/60s
       // ToDo: set checkpoint: write NumConfigs, tdelta, x[i] + bra_x[i]
       printf(" ToDo: write checkpoint data x + bra_x\n");
       t0_solve= t1;
     }
   }

  } // tdelta += increment

  if (Control.output > -1 && myid == 0) {
    printf("\n ========= SOLVE CCM EQUATIONS: COMPLETED! =========\n\n");
  }

  // fclose(terms_fp);    terms_fp=NULL;
  if (myid==0) fclose(fp2);

  // ferase(fp_exp);

  // stat=remove(gs_name);
  // stat=remove(input_name);  
  // stat=remove(configurations);

}

/* called by solve() and directIteration(only_printout)
 * new: x[0]=1 now (Sep05)
 */
double findEnergy(double *x) 
{
  int i, bonds, indices[7]={0};
  double dummy, sum;

  sum=0.0;
  term_rewind(0);

  for(bonds=1;bonds<=Params.bonds;bonds++) {

    for(i=1;;i++) {
      
      term_get(0,indices,&dummy);  // get gse_terms
      
      if (indices[0]<0 || indices[0]>NumberConfigurations) break;
      
      // Joerg Sep05:
      //   according to gse_terms only indices[2..3] can be nonzero
      // gse_term{3}      : (i2,i3) = two single-site configs
      // gse_term{2,4,5,7}: (i2, 0) = two-site neighbour config
      // gse_term{1,6,8}  : ( 0, 0)
      // squareLSUB10 -> B1: (0,1,0,0,0,0,-0.5), B2: (0,27,1,0,0,0,-1.0)
      sum += dummy * x[indices[2]]
                   * x[indices[3]] * delta[bonds][indices[1]];
    }
  }
  return sum;
}
 
// in/out: oldX - startvector + iterationvector
// in/out: dx   - difference oldX(out) - oldX(in)  (only if precondition)
//                used to modify start vector (speed up?) -> removed
// return status 0=ok,  -N = error
int directIteration(int maximumNumIterations, int numConfigs, 
		    double tolerance, int *iterations, double *oldX)
{
  int i, j, bonds, max_ket_i=1, status=0;
  int indices[7]={0};
  double dummy;
  double xi2, xi3, xi4, xi5;
  double *lhs, *rhs, *tmp,    /* Joerg: left-hand side, right-hand side */ 
         min_error, error, newX;
  static double mom = momentum; // set in Constant.h

  error     =  0.0; // JS Nov05
  min_error = 10.0; // JS Apr06, to stop divergence problems LSUB14 chain

  if (myid==0 && Control.output>((SearchAlphaMin)?1:0))
    printf("Direct iteration ket state (tolerance=%g, momentum=%g):\n",
            tolerance, momentum);

  lhs = dvector(0,numConfigs+5); // +5 ???, better +npe for balance?
  rhs = dvector(0,numConfigs+5);
  tmp = dvector(0,numConfigs+5); // for MPI

  for(i=1;i<=numConfigs;i++) {
    lhs[i]=0.0;
    rhs[i]=0.0;
    tmp[i]=0.0;
  }
  tmp[0]=0.0;
  lhs[0]=0.0;
  rhs[0]=0.0; // used as status/error flag, JS0904

  // for comparision old/new versions in case of trouble 
  if (Control.output > ((SearchAlphaMin)?1:0) && myid==0) {
   // for (i=1;i<=NumberConfigurations;i++)
   //   if (oldX[max_ket_i]<oldX[i])
   //            max_ket_i =     i;
    printf(" %5d Eg  %16.12f err %8.1e "
           "x_max %4d % 10.8f\n",
           0, findEnergy(oldX), error,
           max_ket_i, oldX[max_ket_i]);
  }

  
  for((*iterations)=1; (*iterations)<=maximumNumIterations; (*iterations)++) {
    
       // ToDo: call function term_eq(in:x,in:maxex,id=3,out:newex) ???
       // newex[i0]+=terms(x)*maxex[i2]; ???
    term_rewind(1); // .terms

    /* calculate rhs and lhs from input-file */
    for(bonds=1; bonds<=Params.bonds; bonds++) {
      
      /* why loop over bonds? inner loop does not depend on bonds? */
      for(i=1;;i++) {
	
        term_get(1,indices,&dummy);  // .terms
	
	if (indices[0]<0 || indices[0]>NumberConfigurations) break;
	
        dummy *= delta[bonds][indices[1]];
        /* now: oldX[0]=1.0, Joerg Sep05 */
        xi2=oldX[indices[2]]; /* pointing to a configuration index or 0 */
        xi3=oldX[indices[3]];
        xi4=oldX[indices[4]];
        xi5=oldX[indices[5]]; /* 4 indices only for term34 */
    
        /* ToDo: indices are sorted, so that zeros are at end (indices[5]) */
        /* Joerg: we look for terms where indices[2]==indices[0] and rest 0 */
        if (indices[3]==0 && indices[2]==indices[0]) {
          tmp[indices[0]]-=dummy;                 // rarely (diagonal)
          // which terms? ToDo: sort as first!? (remove if construct!?
          // term2-4,54,58
        } else {
          rhs[indices[0]]+=dummy*xi2*xi3*xi4*xi5; // often (offdiagonal)
        }
      }
    }
    // lhs[] is the same after every loop! JS ToDo: speedup if outside
    // only lhs[j],rhs[j] for (((j-1)%npe) == myid) is non zero
    if ((*iterations)==1) {
#if (PARALLEL_FLAG> 0) 
      /* only every npe-th element is nonzero, summarize */
      for(j=1+myid;j<=numConfigs;j+=npe) tmp[j]=1.0/tmp[j];
      // we simply sum, maybe Alltoall + reordering is faster
      MPI_Allreduce(tmp, lhs, numConfigs + 1,
                    MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
#else
      for(j=1;j<=numConfigs;j++) lhs[j]=1.0/tmp[j]; // expensive, do once
#endif
    }

    // only DEBUG (myid1=1 myid2=2 ... myid0=npe-1)
    for(j=1+myid;j<=numConfigs;j+=npe) {  // parallel 1...
      /* old: using tolerance=1e-12 here is not a good idea!? */
      // usually lhs=-0.5,-1.0,... (see term2-4,54,58)
      //                     and delta[]=usr_parameter(tdelta)
      if (fabs(lhs[j])>1e9 && fabs(rhs[j])>1e7) {
        // sometimes (triangle h=2.8 a=(0,-0.5) rhs=2e-4, lhs=1/-9e-16) ???
        // if (rhs[0]==0.0 && Control.output>1) // print only once!
  	  printf("Error in direct iterations: no left-hand side...\n"
            "it= %4d id= %d cfg= %4d rhs= % .4e * lhs= % .4e, abort\n",
             *iterations, myid, j, rhs[j], lhs[j]);
        // rhs range?
        rhs[j]=0;  // last chance?
        // system("sleep 1"); // give other tasks a chance to complain
        // MPI_Abort(MPI_COMM_WORLD, 1);
        rhs[0]=tmp[0]-=1.0; status=-4; // mark as bad!
      }
    }
#if (PARALLEL_FLAG> 0) 
    /* only every npe-th element is nonzero, summarize */
    for(j=1+myid;j<=numConfigs;j+=npe) tmp[j]=rhs[j]*lhs[j];
    // rest of tmp[j(other tasks)] must be 0
    // store the sum of tmp to rhs incl. status lhs[0]
    MPI_Allreduce(tmp, rhs, numConfigs + 1,
                  MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
#else
    for(j=1;j<=numConfigs;j++) rhs[j]*=lhs[j];
#endif
    // now we can use single-task algorithm to go further
    if (rhs[0]<0) { status=-3; break; }

    error=0.0;
    // err=error;
    // max_X_j=0; max_X=0.0; // only DEBUG
    for(j=1;j<=numConfigs;j++) {
      newX = rhs[j];
      error+=fabs(oldX[j]-newX);
      oldX[j]=mom*oldX[j]+(1.0-mom)*newX; // some parallel overhead
      if (fabs(oldX[max_ket_i])
         <fabs(oldX[        j])) max_ket_i=j;
      // lhs[j]=0.0;
      rhs[j]=0.0;
      // ToDo: outside It.: if (fabs(oldX[j])>fabs(max_X)) {max_X=oldX[j]; max_X_j=j;}
    }
    
    /* calling findEnergy() here is a vast of time ??? (Joerg) */
    if ((Control.output > ((SearchAlphaMin)?1:0) && myid==0
      && (((*iterations) % 10)==0 
       || Control.output > 2 || error > 0.0625 || 2*min_error <= error))
#if SearchAlphaMin==0  // reduce output in search mode
      || (error < tolerance && !myid)  // always output if it will break
#endif
     ) {
      // err= sum(fabs(dX[i]),i) (i,dX[i])_for_max(fabs(dX[i])
      printf(" %5d Eg  %16.12f err %8.1e "
             "x_max %4d % 10.8f\n",
	     *iterations, findEnergy(oldX), error,
             max_ket_i, oldX[max_ket_i]);
    }

    if (isnan(error)) {   /* break if +-Inf or NaN */
       if (!myid && Control.output>1)
         printf(" iteration = %4d : energy = %16.10f : "
                "error = %14.12f\nError directIteration: NaN-BREAK\n",
                *iterations, findEnergy(oldX), error);
       status=-4;
       break;
    }

    if (min_error > error) min_error = error;
    if (error < tolerance) break; // convergence

    if ( (error > 4*min_error && error > 1e-4 && (*iterations) > 99)
      || (error > 10 && (*iterations) > 10)
      || status < 0 ) { // divergence? or rhs/lhs=NaN
        // change mom OR set new random startvec OR add small random vec?
        min_error = 1e+11;
//      for(j=1;j<=numConfigs;j++) oldX[j]=0.00001;
        for(j=1;j<=numConfigs;j++) oldX[j]=0.000001;
//        oldX[1]=oldX[2]=0.18;
        // better set mom outside, bad on search and raster else
        // mom=(1.0+3*mom)/4; // more to 1.0, smaller changes on ket_x[x]
        if (!myid && Control.output>1)
          printf("Warning: no convergence, reset x, status=-2, break\n");
        // some times it converges for higher momentum, but often not
        if (status==0) status=-2; 
        break;
    }

  }  // for (*iterations)
  
  if ( (error > 4*min_error && error > 1e-4 && (*iterations) > 99)
    || (error > 10 && (*iterations) > 10) ) status=1; // divergence?


  if ((*iterations) >= maximumNumIterations) {
    if (!myid && Control.output>1) 
       printf("directIteration Error: iterations>=maximumNumIterations\n"); // exit
    status=-6;
  }

  free_dvector(lhs,0,numConfigs+5);
  free_dvector(rhs,0,numConfigs+5);
  free_dvector(tmp,0,numConfigs+5);

  return status;
}

// in:  x[NumConf]
// out: fvec[NumConf], fjac[NumConf][NumConf]
// called by mnewt()=NewtonRawson in num_methods.C (ntrial) times
void usrfun(int n, double *fvec, double **fjac, double *x, int iterations)
{  
  int i, k, bonds;
  int indices[7]={0};
  double dummy;
  double xi2, xi3, xi4, xi5, multiplier;
  static int warn1=0;

  if (Control.output > 3 && myid == 0) {
    printf("usrfun inp.x[]=");    
    for (i=1; i<=NumberConfigurations && i<=6; i++) {
      if (i==6 && NumberConfigurations>6) printf (" ...");
      else printf(" % 10.8f", x[i]);
    }
    printf("\n");
  }
  
  term_rewind(1);
  
  if (sizeof(int *)<=4 && NumberConfigurations>=16384 && myid==0 && warn1==0){
    /* may be the 2D-array is to big for a 32bit-system (nC>=(2GB/8)^.5) */
    printf("Warning usrfun: array may be to big for a 32bit-system\n");
  }
  for(i=1;i<=NumberConfigurations;i++) {    
    fvec[i]=0.0;
    for(k=1;k<=NumberConfigurations;k++) 
      fjac[i][k]=0.0;  /* may fail if nC*nC*double>2GB */
  }  
 
  i=0;
  for(bonds=1;bonds<=Params.bonds;bonds++) {

    for(;;i++) {
      
       // ToDo: call function term_eq(in:x,in:maxex,id=3,out:newex) ???
       // newex[i0]+=terms(x)*maxex[i2]; ???
      term_get(1,indices,&dummy);
      
      if (indices[0] < 0 || indices[0] > NumberConfigurations)
	break;
      
      x[0]=1.0; /* obsolete??? Joerg Sep05 */
      xi2=x[indices[2]];
      xi3=x[indices[3]];
      xi4=x[indices[4]];
      xi5=x[indices[5]]; 

      multiplier = delta[bonds][indices[1]]*dummy;
      
      fvec[indices[0]] += xi2*xi3*xi4*xi5*multiplier;
      
      if (indices[2] > 0) fjac[indices[0]][indices[2]] += xi3*xi4*xi5*multiplier;
      if (indices[3] > 0) fjac[indices[0]][indices[3]] += xi2*xi4*xi5*multiplier;
      if (indices[4] > 0) fjac[indices[0]][indices[4]] += xi2*xi3*xi5*multiplier;
      if (indices[5] > 0) fjac[indices[0]][indices[5]] += xi2*xi3*xi4*multiplier;
       
    }
  }

  if (Control.output > 4 && myid == 0) {
    printf("usrfun (read %d):\n",i);
    for (i=1; i<=NumberConfigurations; i++) {
      if (i==6 && NumberConfigurations>10) printf (" ...\n");
      if (i>5 && i<NumberConfigurations-4) continue;
      for (k=1; k<=NumberConfigurations; k++) {
        if (k<=5 || k>NumberConfigurations-5) printf(" % 10.8f",fjac[i][k]);
        if (k==6 && NumberConfigurations >10) printf (" ...");
      }
      printf(" : % 10.8f \n",fvec[i]);
    }
  }
  
}

/*****************************************************
 *                                                   *
 * Obtain bra state (LU decomp.)                     *
 *                                                   *
 *****************************************************
 *                                                   *
 * x_I        -- ket-state correlatin coefficient.   *
 * \tilde x_I -- bra-state correlatin coefficient.   *
 *                                                   *
 * H = e_0 + sum_I \tilde x_I F_I                    *
 *                                                   *
 *****************************************************
 *                                                   *
 * d H / d (\tilde x_I) =>                           *
 *                                                   *
 * F_I = 0    <= ket-state equations                 *
 *                                                   *
 *****************************************************
 *                                                   *
 * d H / d x_I =>                                    *
 *                                                   *
 * d(e_o)/d(x_I)+sum_J \tilde x_J (d(F_J)/d(x_I)) =0 * 
 *            <= bra-state equations                 *
 *                                                   *
 * use LU decomposition to solve the equations:      *
 *                                                   *
 * F(I,J) (\tilde X(J)) = - e_o(I)                   *
 *                                                   *
 *****************************************************/

void findBraState(double *x, double *bra_x)
{

  /* Determines the bra-state correlation coefficients
     using LU decomposition.
     */

  int      i;
//  int      status;
  void     lubksb(double **a, int n, int *indx, double b[]);
  void     ludcmp(double **a, int n, int *indx, double *d);
  int      *ind;
  double   d;
  double   **bra_eqn;
  // int      length;
  // int      iterations;

  for(i=1;i<=NumberConfigurations;i++) 
    bra_x[i]=0.0; 

  ind=ivector(0,NumberConfigurations+1);
  bra_eqn=dmatrix(0,NumberConfigurations+1,0,NumberConfigurations+1); 

  findBraStateRHS(bra_x,x); // bra_x (-=) GSEterms(x)
  
  // find sum_J \tilde x_J (d(F_J)/d(x_I)) == F(I,J)
  
  findBraEquations(bra_eqn, x);
  
  // LU decomp Numerical Recipes to solve: F(I,J) (\tilde X(J)) = - e_o(I)
  
  ludcmp(bra_eqn,NumberConfigurations,ind,&d);
  lubksb(bra_eqn,NumberConfigurations,ind,bra_x);
  
  free_ivector(ind,0,NumberConfigurations+1);
  free_dmatrix(bra_eqn,0,NumberConfigurations+1,0,NumberConfigurations+1);

}
  

/*****************************************************
 *                                                   *
 * Obtain bra state (direct iteration)               *
 *                                                   *
 *****************************************************
 *                                                   *
 * x_I        -- ket-state correlatin coefficient.   *
 * \tilde x_I -- bra-state correlatin coefficient.   *
 *                                                   *
 * H = e_0 + sum_I \tilde x_I F_I                    *
 *                                                   *
 *****************************************************
 *                                                   *
 * d H / d (\tilde x_I) =>                           *
 *                                                   *
 * F_I = 0    <= ket-state equations                 *
 *                                                   *
 *****************************************************
 *                                                   *
 * d H / d x_I =>                                    *
 *                                                   *
 * d(e_o)/d(x_I)+sum_J \tilde x_J (d(F_J)/d(x_I)) =0 * 
 *            <= bra-state equations                 *
 *                                                   *
 * Use direct iteration to solve:                    *
 *                                                   *
 *    (\tilde X(I)) = { - e_o(I) - \sum_{J != I}     *
 *                F(I,J) (\tilde X(J))}/F(I,I)       *
 *                                                   *
 * for the \tilde X(I)'s.                            *
 *                                                   *
 *****************************************************/

int directIterationBraState(int maximumNumIterations, int numConfigs, 
			    double tolerance, int *iterations, 
			    double *oldX, double *oldBraX)
{

  double error, min_error, mom;
  
  int i, j, bonds, max_bra_i=1;
  int indices[7]={0};
  double dummy[1];
  double xi2, xi3, xi4, xi5,multiplier;
//  double tempSum1, sum1,tempSum2, sum2;
  double subLatticeMag[MAX_UNIT_CELL+1] = {0.0, 0.0};

  double *lhs, *rhs;
#if (PARALLEL_FLAG> 0)
  double *tempLHS, *tempRHS;
  tempLHS=dvector(0,numConfigs-1);
  tempRHS=dvector(0,numConfigs-1);
#endif
  
  error = 0.0;
  min_error= 10.0;
  mom = momentum;
  // mom = 0.7;  // 0.25
  if (myid == 0) printf("Direct iteration bra state (mom=%g):\n",mom);

  lhs=dvector(0,numConfigs+1);
  rhs=dvector(0,numConfigs+1);

  for(i=1;i<=numConfigs;i++) {
    lhs[i]=0.0;
    rhs[i]=0.0;
  }

  for((*iterations)=1;(*iterations)<=maximumNumIterations;(*iterations)++) {
        
#if (PARALLEL_FLAG == 0) 
    findBraStateRHS(rhs,oldX); // rhs (-=) GSEterms(oldX)
#endif

    term_rewind(1);  // .terms

    for(bonds=1;bonds<=Params.bonds;bonds++) {
      
      for(i=1;;i++) {
	
       // ToDo: call function term_eq(in:x,in:maxex,id=3,out:newex) ???
       // newex[i0]+=terms(x)*maxex[i2]; ???
        term_get(1,indices,dummy);
	
	if (indices[0] < 0 || indices[0] > NumberConfigurations)
	  break;

        multiplier = dummy[0] * delta[bonds][indices[1]];

	
	if (indices[2]==0) { // all indices are equal to 0
	  
	  // rhs[indices[0]]+=multiplier;
	  
	} else {
	  
          oldX[0]=1.0;
	  xi2=oldX[indices[2]];
	  xi3=oldX[indices[3]];
	  xi4=oldX[indices[4]];
	  xi5=oldX[indices[5]];  /* only for term34 */

          if (indices[5]>0)
            rhs[indices[5]]-=xi2*xi3*xi4*multiplier*oldBraX[indices[0]];
          if (indices[4]>0)
            rhs[indices[4]]-=xi2*xi3*xi5*multiplier*oldBraX[indices[0]];
          if (indices[3]>0)
            rhs[indices[3]]-=xi2*xi4*xi5*multiplier*oldBraX[indices[0]];

	  if (indices[2]>0) {
	    if (indices[2]==indices[0]  && indices[3] ==0) // [4..5] == 0 too
	      lhs[indices[0]]+=xi3*xi4*xi5*multiplier;
	    else 
	      rhs[indices[2]]-=xi3*xi4*xi5*multiplier*oldBraX[indices[0]];
	  }
	}
      }
    }
      
#if (PARALLEL_FLAG> 0)
    /* ToDo JS06: speedup if rhs/lhs or only error[npe] summed,
     *    changing oldX can be distributed over threads */
    MPI_Allreduce(lhs+1, tempLHS, numConfigs, 
		  MPI_DOUBLE, MPI_SUM,MPI_COMM_WORLD);
    MPI_Allreduce(rhs+1, tempRHS, numConfigs, 
		  MPI_DOUBLE, MPI_SUM,MPI_COMM_WORLD);
    for(i=1;i<=numConfigs;i++) {
      lhs[i]=tempLHS[i-1];
      rhs[i]=tempRHS[i-1];
    }
    findBraStateRHS(rhs,oldX);  // rhs (-=) GSEterms(oldX)
#endif

    error=0.0;
    for(j=1;j<=numConfigs;j++) {
      if (fabs(lhs[j]) < 1e-9) {
	printf("Error in direct iterations bra state: no left-hand side...\n"
          "it= %4d  cfg= %4d  rhs= % .4e lhs= % .4e < 1e-9 ---  Exiting!!\n",
           *iterations, j, rhs[j], lhs[j]);
	break;
      }
      error+=fabs(oldBraX[j]-rhs[j]/lhs[j]);
    }
    if (j<=numConfigs) break;
#if 0  // can generate lot of iterations
    if (error>8*min_error) { // New JS Apr06, 2* => 8* Mar08, ToDo: make bigger?
       mom = 1.0 - 0.5*(1.0 - mom);  /* 1-0.3 -> 1-0.15 -> 1-0.075 -> ... */
       min_error = error;
       if (!myid)
         printf(" It=%d: reduce 1-momentum to %g to ensure convergence\n",
                *iterations, 1.0-mom);
       if (1.0-mom<1e-6) break;
    }
#endif
    for(j=1;j<=numConfigs;j++) {
      oldBraX[j]=mom*oldBraX[j]+(1.0-mom)*rhs[j]/lhs[j];
      if (fabs(oldBraX[max_bra_i])
         <fabs(oldBraX[        j])) max_bra_i=j;
      lhs[j]=0.0;
      rhs[j]=0.0;
    }

    if (Control.output > 0 && myid == 0
       && ((((*iterations) % 20)==0 || Control.output > 2)
       || error < tolerance)) 
      printf(" %5d mag   %14.12f err %8.1e %8.1e bra_max %4d %11.8f\n",
	*iterations, findStagMag(oldX,oldBraX,subLatticeMag),
        error, error-min_error, max_bra_i, oldBraX[max_bra_i]);

    if (error < min_error) min_error = error;
    if (error < tolerance) break;
    // NaN happens for j1-j2-neel j2>0.55 ??? was a bug (Mar09 fixed)?
    if ((isnan(subLatticeMag[1]) || isnan(error))) {
      static int err_nan=0;
      if (++err_nan<10 && myid<10)
        printf("NaN-Error: SublMag= %g err= %g myid= %d (max=9)\n",
          subLatticeMag[1], error, myid);
      if (isnan(error)) ERR("error=NaN");
    }
    if (error > 1e6) { (*iterations)=-1; break; }       // divergence

  }

  if ((*iterations) >= maximumNumIterations) {
    printf("directIterationBraState Error: i>=maximumNumIterations\n");
    ERR(0);
  }

  free_dvector(lhs,0,numConfigs+5);
  free_dvector(rhs,0,numConfigs+5);
#if (PARALLEL_FLAG> 0)
  free_dvector(tempLHS,0,numConfigs-1);
  free_dvector(tempRHS,0,numConfigs-1);    
#endif

  return *iterations;
}

// input:  rhs, x (global: terms, Params.bonds, NumberConfigurations)
// output: rhs    ( -= GSEterms(x) ) 
// --- called from each thread ---
void findBraStateRHS(double *rhs, double *x) 
{
  int i, bonds;
  int indices[7]={0};
  double dummy[1];
  double xi2, xi3, xi4, xi5, multiplier;

       // ToDo: call function term_eq(in:x,in:maxex,id=3,out:newex) ???
       // newex[i0]+=terms(x)*maxex[i2]; ???
  term_rewind(0);

  for(bonds=1;bonds<=Params.bonds;bonds++) {

    for(i=1;;i++) {
      
      term_get(0,indices,dummy); // GSE
      
      if (indices[0] < 0 || indices[0] > NumberConfigurations) break;
      
      multiplier=delta[bonds][indices[1]];

      // ToDo: how often? can be removed?
      if ((indices[2]==0 && indices[3]==0
        && indices[4]==0 && indices[5]==0)) continue;
	
        x[0]=1;
        xi2=x[indices[2]];
	xi3=x[indices[3]];
	xi4=x[indices[4]];
	xi5=x[indices[5]];
	
	if (indices[2] > 0) // ToDo: faster if using dummy rhs[0] ??? 
	  rhs[indices[2]]-= dummy[0]*xi3*xi4*xi5*multiplier;
	if (indices[3] > 0) 
          rhs[indices[3]]-= dummy[0]*xi2*xi4*xi5*multiplier;
	if (indices[4] > 0)
	  rhs[indices[4]]-= dummy[0]*xi2*xi3*xi5*multiplier;
	if (indices[5] > 0)
	  rhs[indices[5]]-= dummy[0]*xi2*xi3*xi4*multiplier;
    }
  }
}

// ToDo: m-indizes are cache unfriendly, change it! JS091107
// work: m[0][x]
// out: m[1..NumCfgs][1..NumCfgs]
// ToDo: melt with usrfun?
void findBraEquations(double **m, double *x) 
{
  /* FIND BRA EQUATIONS: finds the Jacobian in order to determine 
   the bra state equations.
   */

  // find: sum_J \tilde x_J (d(F_J)/d(x_I))
  
  int i,k,bonds;
  int indices[7]={0};
  double dummy[1];
  double xi2, xi3, xi4, xi5, multiplier;
  
  term_rewind(1);
  
  for(i=1;i<=NumberConfigurations;i++)    
    for(k=1;k<=NumberConfigurations;k++) 
      m[i][k]=0.0;

  for(bonds=1;bonds<=Params.bonds;bonds++) {
    
    for(i=1;;i++) {
      
      term_get(1,indices,dummy); // .terms
      
      if (indices[0] < 0 || indices[0] > NumberConfigurations)
	break;
      
      multiplier=dummy[0]*delta[bonds][indices[1]];

      // check that one of indices[2..5] is not zero 
      if (indices[2]==0) continue; // rarely!? dont needed 
      
      x[0]=1.0;  // dummy factor 1 does not change anything
      xi2=x[indices[2]];
      xi3=x[indices[3]];
      xi4=x[indices[4]];
      xi5=x[indices[5]];
      
      if (indices[2] > 0)
        m[indices[2]][indices[0]]+= xi3*xi4*xi5*multiplier;
      if (indices[3] > 0)
        m[indices[3]][indices[0]]+= xi2*xi4*xi5*multiplier;
      if (indices[4] > 0)
        m[indices[4]][indices[0]]+= xi2*xi3*xi5*multiplier;
      if (indices[5] > 0)
        m[indices[5]][indices[0]]+= xi2*xi3*xi4*multiplier;
    }
  }
}

/*****************************************************
 *                                                   *
 * Name: Staggered Magnetisation Routine.            *
 *                                                   *
 * Determines the lattice magnetisation once bra-    *
 * and ket-state equations have been solved.         *
 *****************************************************/

double findStagMag(double *x, double *bra_x, double *subLatticeMag)
{

  /* FIND STAG MAG: obtains the staggered magnetisation.
     */

  int      i, j;
  double   stagMag;
  int      length;
  int      (*configuration)[5];
  int      subTerm[MAX_UNIT_CELL+1];

  configuration = (int (*)[5]) malloc( (MAX_LSUBM+1)*5*sizeof(int) );
  if (!configuration) { printf("ERROR findStagMag: malloc\n"); return 0; }
  /* ToDo JS0911: how to handle error in MPI case? */

  // first find the \sum_i < \tilde \Psi | s_i^z | \Psi > bit.

  stagMag=0.0; 
  
  for(i=1;i<=UnitCell.number;i++) {
    subLatticeMag[i] = 0.5*UnitCell.s2[i];
    stagMag         += 0.5*UnitCell.s2[i];
  }
  
  stagMag=stagMag/((double) UnitCell.number);

  for(i=1;i<=NumberConfigurations;i++) {

    length=findClusterFromDescriptor(spinConfigurations[i].descriptor,
				     configuration);

    stagMag -= length*bra_x[i]*x[i];    
    sublattice_term(length,configuration,subTerm);
    
    for(j=1;j<=UnitCell.number;j++) 
      subLatticeMag[j] -= subTerm[j] * bra_x[i] * x[i]
                        * UnitCell.number / LatticeSymmetries.number;
  }

  free(configuration);
  return stagMag;
}

void sublattice_term(int length, int (*configuration)[5], int *subTerm)
{
  
  /* 
     This routine finds contribution to the sublattice magnetisation
     from a given cluster, namely:
     
     F(m) . S(+)(m) 
  */
  int i, j;
  static int tempConfiguration[MAX_LSUBM+1][5];
  
  for(i=1;i<=UnitCell.number;i++) subTerm[i]=0;

  // find symmetry related clusters
  for(j=1;j<=LatticeSymmetries.number;j++) { 
    /* ToDo: configuration modified ?? */
    symmetryGroup(j,length,configuration,tempConfiguration);
    for(i=1;i<=length;i++) { subTerm[tempConfiguration[i][4]]++; }
  } 

  return;
}

// new code to find the generalised expectation values.

void findGenExp(double *x, double *bra_x, double *g_exp, double *alpha)
{

  /* FIND GENERALISED EXPECTATION VALUES: obtains the staggered magnetisation.
     */

  int      i,j,index,term,offs;
  int      (*configuration)[5];
  int      expect,error;
  int	   indices[7]={0};
  double   dummy[1],multiplier;
  double   xi2, xi3, xi4, xi5,bx;

  configuration = (int (*)[5]) malloc( (MAX_LSUBM+1)*5*sizeof(int) );
  if (!configuration) { printf("ERROR findStagMag: malloc\n"); exit(1); }
  /* ToDo JS0911: how to handle error in MPI case? */
    
  // JS0902: for parallel, we store GSE terms temporary 
  // g_exp[1..nExpect] final result, g_exp[nExpect+2..2*nExpect] GSE terms
  for(i=0;i<=2*Params.nExpect;i++) g_exp[i]=0.0;

       // ToDo: call function term_eq(in:x,in:maxex,id=3,out:newex) ???
       // newex[i0]+=terms(x)*maxex[i2]; ???
  term_rewind(2); // term_open(2,"r"); Jun08

  for(expect=1;expect<=Params.nExpect;expect++) {

    for (j=1;j<=2;j++) { // GSE terms + hh terms
      for(i=1;;i++) {  // sum all terms

        error=term_get(2,indices,dummy); /* .exp_terms */

        if (indices[0] < 0 || indices[0] > NumberConfigurations) break;

        index=indices[0];
        // the index for the bra-state correlation coefficient.
	// if this is zero then it's one of the "gs" like terms.
	
        // ToDo: use bra_x[0]=1.0
	term=indices[1];
	bx=bra_x[index];
	if (index < 1) bx=1.0;
	
	xi2=x[indices[2]];
	xi3=x[indices[3]];
	xi4=x[indices[4]];
	xi5=x[indices[5]];

        multiplier=1.0;

#if UseExpAlpha
         // arg2 is term13 (range=1..13) 2008-10-01
        // DJJF0902: usr_exp_angle can be different from usr_angle
        //           exsample: lattice magnetization of triangular lattice
        multiplier=usr_exp_angle(expect,map62[term],alpha,UseAlpha);
	// bad: multiplier=usr_exp_angle(expect,term,alpha,UseAlpha);
#endif
			
#if DEBUG // keep code small for CPU Icache
	if (Control.output > 3 && myid ==0) 
	  printf("expect=%d : config=%d : term =%d : (%d,%d,%d,%d) : mult1=%10.6f : mult2=%10.6f : g_exp=%10.6f\n",
	         expect,indices[0],indices[1],indices[2],indices[3],indices[4],indices[5],
		 dummy[0],multiplier,g_exp[expect]);
#endif

        g_exp[expect] += bx*xi2*xi3*xi4*xi5*dummy[0]*multiplier;
			
      } // i=1...
      if (j==1) { // GSE
        offs=Params.nExpect;
        g_exp[offs+expect]=g_exp[expect];
        g_exp[     expect]=0.0;
      }
    } // j=1..2

  } // expect=1..nExpect
#if (PARALLEL_FLAG> 0)
  {  int j;
     double *tmp_exp=(double *)malloc((Params.nExpect+1)*sizeof(double));
     if (!tmp_exp) MPI_Abort(MPI_COMM_WORLD, 1);
     for(j=1;j<=Params.nExpect;j++) tmp_exp[j]=g_exp[j];
     MPI_Allreduce(&tmp_exp[1], &g_exp[1], Params.nExpect,
                     MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
     free(tmp_exp);
  }
#endif
  for(j=1;j<=Params.nExpect;j++) g_exp[j]+=g_exp[Params.nExpect+j];

  // fclose(fp_exp); Jun08
  free(configuration);
  return;
}

// DJJF Jun08: new code to find the excitation energy values.
// Joerg: add MPI functionality

void findExcitations(double *x, double *bra_x, double *alpha, double *ee)
{

  /* Find the lowest lying excitation energies (ee) from the CCM excited-state equations.

     Solved as an eigenvalue problem using the eispack routine rg_().
	 
	 Written by Damian JJ Farnell: 27 June 2008. 
	 
	 Copyright: Damian JJ Farnell (June 2008). All rights reserved.
     */

  int      i,j,index,term,bonds; // unused? ,k;
  int      term_get_error;
  int	   indices[7]={0};
  double   dummy[1];
  double   xi3, xi4, xi5;
  double   multiplier1; // ,multiplier2;
  // double   factorial[100];  
  // int length1, length2;
  // int (*tempConfig)[5]=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));  
//  int term2;

  // factorial[0]=1.0;	
  // factorial[1]=1.0;
  // for(i=2;i<100;i++) factorial[i]=factorial[i-1]*((double) i);
	
//  double ee=0.0;
    
  if (solutionStrategy==1) { // Newton-Raphson(?) (serial, but nC^2 memory)
    // new code written 19 December 08 DJJF
    //  strategy=0 - direct iteration excited-state CCM (parallel)

    // now open the excitations CCM equations file.

       // ToDo: call function term_eq(in:x,in:maxex,id=3,out:newex) ???
       // newex[i0]+=terms(x)*maxex[i2]; ???
    term_rewind(3);	// .exc_terms
      
    // find the initial v
      
    static double *m=NULL;  // matrix
    // if (NumberExcitedConfigurations > 1) 
    // Joerg Jun08: needs much memory cfg^2, ToDo: split into threads!
    {
      if (!m) // malloc once, never free
      m=dvector(0,(NumberExcitedConfigurations)
                 *(NumberExcitedConfigurations));
      if (!m) { printf("ERROR: ecs_matrix=malloc(numExcCfg^2)\n"); return; }
      for (i=0;i<NumberExcitedConfigurations
                *NumberExcitedConfigurations;i++) m[i]=0.0;
    }
    *ee=0.0;
      
    for(bonds=1;bonds<=Params.bonds;bonds++) {
      for(i=1;;i++) {
        term_get_error=term_get(3,indices,dummy); // .exc_terms
        if (indices[0]<0) break;
#if DEBUG // keep code small for CPU Icache
        if (indices[0] <= 0 || indices[0] > NumberExcitedConfigurations) {
          printf("Something funny: index[0] =%d which is <= 0 or > num exc configs!\n",indices[0]);
          exit(0);
        }
#endif
        x[0]=1.0; // JS0902: trick, replacing if constructs

        index=indices[0]; // the index for the excited correlation coefficient.
                          // This is never zero for excited states	
                            
        term =indices[1]; // this is the term number (of 81) in the excited-state equations  

        xi3=x[indices[3]];    // a CCM correlation coefficient
        xi4=x[indices[4]];    // a CCM correlation coefficient
        xi5=x[indices[5]];    // a CCM correlation coefficient
                            
        // term2=mapping[term];  // map 1..81 => 1..62 [ => 1..13]
        // Joerg Jun08: mapping done in obtainExc... now
        // term2=term;

        multiplier1 = delta81[bonds][term];
#if 0 && UseAlpha  // ToDo: error, twice usr_angle?
        multiplier1 *= usr_angle(bonds/*expect*/,term,alpha,UseAlpha);
#endif
        {
          // x[2] <- excitation coefficients
          // x[3] to x[5] <= g-s ket-state coefficients (NO PARTICULAR ORDER)
          index=(indices[0]-1)*NumberExcitedConfigurations
                +indices[2]-1;
#if DEBUG
          if (index >=  NumberExcitedConfigurations
                       *NumberExcitedConfigurations) {
             printf("Index > ne*ne!\nExiting..\n\n"); ERR(1);
          }
#endif
          m[index] += xi3*xi4*xi5*dummy[0]*multiplier1;
        }
#if DEBUG
        if (Control.output > 2) 
          printf("Bond %d cfg %d t%02d.%02d : %d %d %d %d m1=%10.6f m2=%10.6f tot=%10.6f\n",
              bonds, indices[0], map81[indices[1]], indices[1], indices[2],
              indices[3], indices[4], indices[5],
              dummy[0], multiplier1, xi3*xi4*xi5*dummy[0]*multiplier1);
#endif
      }
    }

    // fclose(fp_exc);

#if (PARALLEL_FLAG> 0)
    // if (NumberExcitedConfigurations > 1) 
    {
      static double *m2=NULL;
      if (!m2)
      m2=dvector(0,(NumberExcitedConfigurations)
                  *(NumberExcitedConfigurations));
      for (i=0;i<NumberExcitedConfigurations
                *NumberExcitedConfigurations;i++) m2[i]=m[i];
      MPI_Allreduce(m2, m, NumberExcitedConfigurations
                          *NumberExcitedConfigurations,
                    MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
      free_dvector(m2,0,NumberExcitedConfigurations
                       *NumberExcitedConfigurations);
      m2=NULL;
      // now we can use single-task algorithm to go further
    }
#endif

    // non-parallel
    // order1=(long) NumberExcitedConfigurations; => excMatrixSize
    *ee=m[0];
    if (NumberExcitedConfigurations>1) {  // else *ee=m[0];

      int imin=0, iminr=0, imax=0;
      double *wi,*wr,*z,*fv1;
      long int *iv1;

      // ToDo: Jun08, alpha version had a memory leak!
      wi =dvector(0,(NumberExcitedConfigurations)); 
      wr =dvector(0,(NumberExcitedConfigurations)); 
      iv1=(int long *)malloc(2*NumberExcitedConfigurations*sizeof(long int));
      if (!iv1) ERR("malloc iv1[2*NumExcCfgs]");
      fv1=dvector(0,(2*NumberExcitedConfigurations)); 
      z  =dvector(0,(2*NumberExcitedConfigurations));
     
      long int order1=0;
      long int    matz=0;
      long int    ierr=0;
      int stat;

      order1=(long) NumberExcitedConfigurations;

  // DJJF test: matrix LSUB4 Heisenberg square from old CCM code
  // Joerg: hmm, I have other results for LSUB4 Heisenberg square
  /* 6*6 matrix:
   *m= { 1.122, -12.000, -12.000, -24.000, -24.000, -24.000, 
        -0.048,   4.982,   0.131,  -0.046,  -0.092,  -0.033, 
         0.005,   0.075,   5.985,  -0.006,  -0.052,  -0.350, 
        -0.038,  -0.023,  -0.017,   5.270,  -0.367,  -0.102, 
         0.004,   0.000,  -0.020,  -0.160,   5.963,  -0.195, 
         0.009,   0.003,  -0.175,  -0.005,  -0.195,   5.878};  // OK 
  */
  // DJJF: Gives the correct value of "Excited-State Energy = 0.8497504284"
  // JS:   but LSUB4 value and eigenvalue(m) is: 0.851867601151 !!! (see papers)
  //       inaccuracy? because we only have %.3f here?

      if ((Control.output>3 || order1<13) && myid==0) {
        printf("\nCCM Excited-State Eigenvalue Matrix To Be Diagonalised = \n");
        for (i=0; i<order1; i++) {    
          for (j=0; j<order1; j++) {
            if (order1<6)
              printf(" % 11.9f",m[i*NumberExcitedConfigurations+j]);
            else {
              if (order1<8)
                printf(" % 9.7f",m[i*NumberExcitedConfigurations+j]);
              else
                printf(" % 8.6f",m[i*NumberExcitedConfigurations+j]);
            }
          }
          printf("\n");
        }
        printf("\n");
      }

      stat=rg_(&order1 , &order1 , m ,wr , wi, &matz, z, iv1, fv1, &ierr);
        // eispack eigenvalue routine
                  
      if (ierr !=0) 
        printf("Ierror in solve excitations = %d\n",(int) ierr);
     
      *ee=1e10; // ee contains the lowest-valued eigenvalue
      for (i=0; i<order1; i++) {
        if (Control.output > 2 && myid==0 && i<Control.output*100) {
          printf("Excited-state Eigenvalue %2d = ( %16.12f , %16.12f ) \n",
                 i+1, wr[i], wi[i]);
        }
        // ToDo: min. absolut or min reel
        if (fabs(wi[i])<1e-6)               // skip complex values
        if (wr[i] < wr[imin ]) { imin= i; }
        if (wr[i] < wr[iminr]) { iminr=i; } // include complex values
        if (wr[i] > wr[imax ]) { imax= i; }
      }
      *ee = wr[imin];
      if (myid==0) {
        printf("Min Excited-state Eigenvalue %3d = ( %16.12f , %16.12f )\n",
               imin, wr[imin], wi[imin]);
        if (imin!=iminr)
        printf("Min Excited-state Eigenval.? %3d = ( %16.12f , %16.12f )\n",
               iminr, wr[iminr], wi[iminr]);
        printf("Max Excited-state Eigenvalue %3d = ( %16.12f , %16.12f )\n",
               imax, wr[imax], wi[imax]);
        if (fabs(wi[iminr])>1e-6) printf("Warning: complex minGAP\n");
      }
      free_dvector(wi, 0,(NumberExcitedConfigurations));
      free_dvector(wr, 0,(NumberExcitedConfigurations));
      free(iv1);
      free_dvector(fv1,0,(2*NumberExcitedConfigurations));
      free_dvector(z,  0,(2*NumberExcitedConfigurations));
    } // NumberExcitedConfigurations>1

  } else {  // solutionStrategy != 1
    //  strategy=0 - direct iteration excited-state CCM (parallel)

    // new code written 19 December 08 DJJF
    //  - direct iteration excited-state CCM
    // shifted power method
    // this should "track" the maximum and minimum eigenvalue solution

    double *newex,error;
    double newee=1.0; // JS08: was uninitialized
    int maximumNumIterations, iterations;
    double maxee=0.0; // 12.07; // JS091106 empirical
    time_t t0, t1; /* measure time in seconds */
   
    newex=dvector(0,(NumberExcitedConfigurations));  /* len=NumEx+1 */
    if (!newex) { printf("ERROR newex malloc\n"); exit(1); }
    maximumNumIterations=(int) MAX_NUM_ITERATIONS;
    //	maximumNumIterations=50000;
#if (PARALLEL_FLAG> 0) 
    static double *m2=NULL;
    if (!m2) m2=dvector(0,NumberExcitedConfigurations); /* 0..NumEx */
#endif

    // added by JS 2009-07: maxex[2] is growing endless for xxzTriangle LSUB2
//    for(i=1;i<=NumberExcitedConfigurations;i++)  {
//       ex[i]=0.01; 
//       maxex[i]=0.01;
//    }

    // first find the maximum eigenvalue
  
    time(&t0);
    *ee=1.0e10;

    for(iterations=1; iterations<=maximumNumIterations; iterations++) {

       // ToDo: call function term_eq(in:x,in:maxex,id=3,out:newex)
       // newex[i0]+=terms(x)*maxex[i2];
 
       // ToDo: call function term_eq(in:x,in:maxex,id=3,out:newex) ???
       // newex[i0]+=terms(x)*maxex[i2]; ???
       term_rewind(3);
       /* Jan09 JS: [0] unused, [1..NumEx] used */
       for(j=1;j<=NumberExcitedConfigurations;j++) newex[j]=0.0;
       
       for(bonds=1;bonds<=Params.bonds;bonds++) {
         for(i=1;;i++) {
         
         term_get_error=term_get(3,indices,dummy); // .exc_terms
         
         if (indices[0]<0) break;
#if DEBUG // keep code small for CPU Icache
         if (indices[0] <= 0 || indices[0] > NumberExcitedConfigurations) {
           printf("Something funny: index[0] =%d which is <= 0 or > num exc configs!\n",indices[0]);
           exit(0);
         }
#endif
         index=indices[0]; // the index for the excited correlation coefficient.
                 // This is never zero for excited states	
                 
         term=indices[1]; // this is the term number (of 81) in the excited-state equations  

         x[0]= 1.0; // avoid if(index[3..5]<1) xi3..5=1.0
         xi3 = x[indices[3]];    // a CCM correlation coefficient
         xi4 = x[indices[4]];    // a CCM correlation coefficient
         xi5 = x[indices[5]];    // a CCM correlation coefficient

         // map82[term];  // map 1..81 => 1..62 [ => 1..13]
         // Joerg Jun08: map82[term82] done in obtainExc... now
         // term2=term;

         multiplier1 = delta81[bonds][term];
#if 0 && UseAlpha  // ToDo: bug? 2 times usr_angle?
         multiplier1 *= usr_angle(bonds,term,alpha,UseAlpha);
#endif
         newex[indices[0]] += xi3*xi4*xi5*dummy[0]
                           *multiplier1*maxex[indices[2]];
       } // end read terms
     }  // end bonds

     /* JS09: is this the right position for parallel code? sum newex[]
      *   and is it the right code?
      * LC-6-6 s=1 Gap(?) =      12.07000109 
      */
#if (PARALLEL_FLAG> 0) 
      for (i=1;i<=NumberExcitedConfigurations;i++) m2[i]=newex[i];
      MPI_Allreduce(m2+1, newex+1, NumberExcitedConfigurations,
                    MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
      // free_dvector(m2,0,NumberExcitedConfigurations); m2=NULL;
#endif
// JS091106: maxee was 0 
//     // subtract maxEV, so we can itereate to minE-maxE as biggest |EV|
      for(j=1;j<=NumberExcitedConfigurations;j++) newex[j]-=maxee*maxex[j];

      // done: newex = ExcEqMatrix(x) * maxex

      // JS 2009-07 the matrix may be diagonal, than first nonzero is not best
      // JS091105: better go around 2 times max/search?
#if 0
      // search first non-zero value
      for(j=1;j<=NumberExcitedConfigurations;j++) {
        if (fabs(newex[j])>0.0000001) { newee=newex[j]; break; }
      }
#else
      double sum_a=0, min_e=0; // DBG
      newee=0;
#if EXPERIMENTAL & 4
      newee=1e10;
#endif
      for(j=1;j<=NumberExcitedConfigurations;j++) {
#if EXPERIMENTAL & 4
        if (newex[j]<newee) newee=newex[j];  // min. Eigenvalue exc. State
#else 
        if (newex[j]>newee) newee=newex[j];  // max. Eigenvalue exc. State
#endif
        if (newex[j]>min_e*maxex[j]) min_e=newex[j]/maxex[j];
      }
#endif
      for(j=1;j<=NumberExcitedConfigurations;j++)
        maxex[j]=0.5*newex[j]/newee+0.5*maxex[j]; // new = new/2 + old/2
//        maxex[j]=(1-momentum)*newex[j]/newee+momentum*maxex[j]; // sq8 better?
//        maxex[j]=newex[j]/newee;
newee+=maxee;  // shift back energy JS091106
#if EXPIRIMENTAL & 4
if (maxee==0.0) maxee=min_e;
#endif
if(!myid) if (0==iterations%20 || iterations<10)
  printf(" min_e=%12.8f sum_a=%12.8f maxee %12.8f %12.8f\n",
           min_e, sum_a, maxee, newee);
          
      error=fabs(*ee-newee);
      *ee=newee;

      if (!myid) {
        if (Control.output>0 && 1==iterations) 
          printf("Direct Iteration Max. Eigenvalue Excited State:\n");
        if (Control.output > 2
        || (Control.output > 0 && (0==iterations%20 || iterations<5))
        || error < ((double) TOLERANCE)) 
          printf("Max. Eigenvalue Excited State: It= %4d  Gap= %12.10f  Err= %.2e\n",
            iterations, newee, error);
        if (Control.output > 2) {
          printf("maxex[]=");
          for(j=1; j<=NumberExcitedConfigurations && j<=6; j++)
              printf(" %9.6f", maxex[j]); // ToDo: range?
          printf("\n");
          printf("newex[]=");
          for(j=1; j<=NumberExcitedConfigurations && j<=6; j++)
              printf(" %9.6f", newex[j]); // ToDo: range?
          printf("\n");
        }
      }
              
      if (error < ((double) TOLERANCE)) break;

    } // iterations loop

#if EXPERIMENTAL & 4
return;
#endif
    maxee=*ee;
    *ee=1.0e10;

    if (iterations < maximumNumIterations)  // JS 2009-07 we cant trust value if not converged
                                            // JS 2009-07: skip 2nd iteration if first was bad 
    for(iterations=1; iterations<=maximumNumIterations; iterations++) {

     // ToDo: call function term_eq(in:x,in:ex,id=3,out:newex)
     // newex[i0]+=terms(x)*ex[i2];

     term_rewind(3);
     for(j=1;j<=NumberExcitedConfigurations;j++) newex[j]=0.0;
     
     for(bonds=1;bonds<=Params.bonds;bonds++) {
       for(i=1;;i++) {
        
         term_get_error=term_get(3,indices,dummy); // .exc_terms
        
         if (indices[0]<0) break;
#if DEBUG // keep code small for CPU Icache              
         if (indices[0] <= 0 || indices[0] > NumberExcitedConfigurations) {
           printf("Something funny: index[0] =%d which is <= 0 or > num exc configs!\n",indices[0]);
           exit(0);
         }
#endif

         index=indices[0]; // the index for the excited correlation coefficient.
                // This is never zero for excited states	
                
         term=indices[1]; // this is the term number (of 81) in the excited-state equations  

         x[0]= 1.0; // JS0902: avoid if(indices[3..5] < 1) xi3..5 = 1.0;
         xi3 = x[indices[3]];    // a CCM correlation coefficient
         xi4 = x[indices[4]];    // a CCM correlation coefficient
         xi5 = x[indices[5]];    // a CCM correlation coefficient

         // term2=mapping[term];  // map 1..81 => 1..62 [ => 1..13]
         // Joerg Jun08: mapping done in obtainExc... now
         // term2=term;

         multiplier1 = delta81[bonds][term];
#if 0 && UseAlpha  // ??? JS: set (const?) delta outside this loop!? bug: 2times?
         multiplier1 *= usr_angle(bonds,term,alpha,UseAlpha);
#endif
         newex[indices[0]]+=xi3*xi4*xi5*dummy[0]*multiplier1*ex[indices[2]];
                        
       }
     }

     /* JS09: is this the right position for parallel code? sum newex[]
      *   and is it the right code?
      * LC-6-6 s=1 Gap    =       0.68253667 
      * JS0910 ToDo: takes a lot of time to get the maxEV,
      *       better to get |minEntry|+|maxEntry| in one run
      *       and use as maxex instead of maxEV (maxEV and minEV could be
      *       computed in the same loop
      */
#if (PARALLEL_FLAG> 0) 
     for (i=1;i<=NumberExcitedConfigurations;i++) m2[i]=newex[i];
     MPI_Allreduce(m2+1, newex+1, NumberExcitedConfigurations,
                  MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
#endif
     // subtract maxEV, so we can itereate to minE-maxE as biggest |EV|
     for(j=1;j<=NumberExcitedConfigurations;j++) newex[j]-=maxee*ex[j];

     // JS 2009-07 the matrix may be diagonal, than first nonzero is not best
     newee = 0.0; // added JS 2009-07 xxzSquare LSUB2 restr=3 nEC=1 newex[1]=0
                  // we only have one excitation? (ToDo: DJJF check that!)
#if 0
     for(j=1;j<=NumberExcitedConfigurations;j++) {
       if (fabs(newex[j])>0.0000001) { newee=newex[j]; break; }
     }
#else
     newee=maxee; // search lowest
     for(j=1;j<=NumberExcitedConfigurations;j++) {
       if (newex[j]<newee) newee=newex[j];
     }
#endif

     if (newee) // added JS 2009-07 avoid div 0
     for(j=1;j<=NumberExcitedConfigurations;j++)
//        ex[j]=0.5*newex[j]/newee+0.5*ex[j];
        ex[j]=newex[j]/newee;
             
     newee+=maxee;  // shift back energy

     error=fabs(*ee-newee);
     *ee=newee;
      
     if (!myid) {
        if (Control.output > 0 && iterations==1) 
             printf("Direction Iteration Min. Excited State:\n");
        if (Control.output > 2
        || (Control.output > 0 && iterations%20==0)
        || error < ((double) TOLERANCE)) 
             printf("Min. Eigenvalue Excited State: It= %4d  Gap= %12.10f  Err= %.2e\n",
                iterations, newee, error);
        if (Control.output > 2) {
          printf("   ex[]=");
          for(j=1;j<=NumberExcitedConfigurations && j<=6;j++)
            printf(" %9.6f",ex[j]);
          printf("\n");
          printf("newex[]=");
          for(j=1;j<=NumberExcitedConfigurations && j<=6;j++) 
            printf(" %9.6f",newex[j]);
          printf("\n");
        }
      }
             
      if (error < ((double) TOLERANCE)) break;

    } // iterations loop

    time(&t1);
    if (!myid)  //                       MinGAP  - MaxGAP (TripletGAP)
      printf("ExcitedState It= %4d  Gap= %12.10f - %12.10f Gap.t[min]= %6.2f\n",
                 iterations, newee, maxee, (t1-t0)/60.);

    if (iterations>=maximumNumIterations) { // JS 2009-07 we cant trust value if not converged
      if (!myid) printf("WARNING: no convergence, resetting values to 0\n");
      *ee=0;
    }


    free_dvector(newex, 0,(NumberExcitedConfigurations)); // remember to free the memory used.
#if (PARALLEL_FLAG> 0) 
    free_dvector(m2,0,NumberExcitedConfigurations); m2=NULL;
#endif

  }  // if (solutionStrategy == 1) ... else ...DI...

  return;
}
